#include "evtid.h"
#include <sys/epoll.h>
#include <unistd.h>
#include <errno.h>
#include <stdlib.h>
#include <assert.h>
#include <fcntl.h>
#include <stdint.h>

// this structure describes the manager of events
typedef struct evd_ctrl_t evd_ctrl_t;

struct evd_ctrl_t{
    unsigned int SocksCount;
#ifndef NDEBUG
    unsigned int readers;
    unsigned int writers;
#endif
    int epfd;
    llist_t ready;
};


static evd_ctrl_t Evd = { .epfd = -1 };

#ifndef NDEBUG
#define INC_USR(type) (Evd.type++)
#define DEC_USR(type) (Evd.type--)
#else
#define INC_USR(type) ((void) 0)
#define DEC_USR(type) ((void) 0)
#endif

static void
initiate_underlying_system(void)
{
    memset(&Evd, 0, sizeof(Evd));
    Evd.epfd = epoll_create(100);
    if(Evd.epfd < 0)
        error(EXIT_FAILURE, errno, "Can't initiate epoll");
    fcntl(Evd.epfd, F_SETFD, FD_CLOEXEC);
    llist_init(&Evd.ready);
}

static void
shutdown_underlying_system(void)
{
    assert(Evd.SocksCount == 0);
    close(Evd.epfd);
    Evd.epfd = -1;
}

int
evtid_in_use(void)
{
    return Evd.epfd >= 0;
}

static void
modify_fd_flags(evd_t * evd, uint32_t flag)
{
    int chk;
    struct epoll_event dsc;

    dsc.data.ptr = evd;
    dsc.events = flag | EPOLLONESHOT;
    chk = epoll_ctl(Evd.epfd, EPOLL_CTL_MOD, evd->d, &dsc);
    if(chk != 0) error(EXIT_FAILURE, errno, "Failed to change events for descr %d", evd->d);
}

static void 
call_real_manager(int wait_time)
{
    int chk;
    struct epoll_event events[Evd.SocksCount + 1];

    // we must have either writers, or readers or wait time. Otherwise
    // we'll end up in epoll forever (or very long time)!
    assert(Evd.readers != 0 || Evd.writers != 0 || wait_time >= 0);

    chk = epoll_wait(Evd.epfd, events, Evd.SocksCount + 1, wait_time);
    if(chk < 0)
    {
        if(errno != EINTR) error(EXIT_FAILURE, errno, "Unexpected error from epoll");
    }
    else if(chk > 0)
    {
        int i;
        for(i = 0; i < chk; i++)
        {
            register struct epoll_event *inf;
            register evd_internal_t * slot;
            evd_t * evd;

            inf = events + i;
            evd = inf->data.ptr;
            slot = &evd->_data;

            // there's a catch. If I request read+write with ONESHOT, then when
            // I get only read or only write, then I never get the other one.
            // so here I have to rearm the check if I didn't get both requests
            // cleared
            if( inf->events & (EPOLLIN|EPOLLHUP|EPOLLERR) )
            {
                SET_READ_READY(slot); 
                if(CHKS_READ(slot))
                {
                    CLR_READ_CHK(slot);
                    DEC_USR(readers);
                }
            }
            if( inf->events & (EPOLLOUT|EPOLLHUP|EPOLLERR) )
            {
                SET_WRITE_READY(slot); 
                if(CHKS_WRITE(slot))
                {
                    CLR_WRITE_CHK(slot);
                    DEC_USR(writers);
                }
            }
            if(llist_empty(&slot->l)) llist_add(&Evd.ready, &slot->l);

            if(CHKS_WRITE(slot))
                modify_fd_flags(evd, EPOLLOUT);
            else if(CHKS_READ(slot))
                modify_fd_flags(evd, EPOLLIN);
        }
    }
}

void
evd_watch(evd_t * evd)
{
    evd_internal_t * slot = &evd->_data;
    int chk;
    struct epoll_event dsc;

    assert(!IS_WATCHED(slot) && evd->d >= 0);

    SET_WATCHED(slot);

    // I request read readinnes, it won't be reported if the user does not allow it
    dsc.events = EPOLLIN|EPOLLONESHOT;
    dsc.data.ptr = evd;

    chk = epoll_ctl(Evd.epfd, EPOLL_CTL_ADD, evd->d, &dsc);
    if(chk != 0) error(EXIT_FAILURE, errno, "Failed to add descriptor to epoll");

    Evd.SocksCount++;
}

void 
evd_nowatch(evd_t * evd)
{
    register evd_internal_t * slot = &evd->_data;
    int chk;
    struct epoll_event ev = {0};

    assert(IS_WATCHED(slot));
    assert(Evd.SocksCount != 0);

    chk = epoll_ctl(Evd.epfd, EPOLL_CTL_DEL, evd->d, &ev);
    if(chk != 0) error(EXIT_FAILURE, errno, "Failed to remove socket from epoll");
    if(!llist_empty(&slot->l)) llist_drop(&slot->l);
#ifndef NDEBUG
    if(CHKS_WRITE(slot)) DEC_USR(writers);
    if(CHKS_READ(slot)) DEC_USR(readers);
#endif

    slot->flags = 0;

    Evd.SocksCount--;
}

// this function requests that read-readinnes is checked
void
evd_input_exhausted(evd_t * evd)
{
    register evd_internal_t * slot = &evd->_data;
    uint32_t flag;

    assert(IS_WATCHED(slot));

    if(CHKS_READ(slot)) return;

    // remove read readinnes indication
    CLR_READ_READY(slot);
    SET_READ_CHK(slot);
    INC_USR(readers);

    flag = EPOLLIN;
    if(CHKS_WRITE(slot)) flag |= EPOLLOUT;
    modify_fd_flags(evd, flag);
}

// this function requests that write-readinnes is checked
void 
evd_check_waction(evd_t * evd)
{
    register evd_internal_t * slot = &evd->_data;
    uint32_t flag;

    assert(IS_WATCHED(slot));

    // if the check is requested already, then we don't do anything else
    if(CHKS_WRITE(slot)) return;

    // just in case, clear write-readinness flag.
    CLR_WRITE_READY(slot);

    flag = EPOLLOUT;
    if(CHKS_READ(slot)) flag |= EPOLLIN;
    modify_fd_flags(evd, flag);

    SET_WRITE_CHK(slot);
    INC_USR(writers);
}

