#include "evtid.h"
#include "xalloc.h"
#include <string.h>
#include <stddef.h>
#include <sys/poll.h>
#include <stdlib.h>
#include <errno.h>
#include <error.h>
#include <assert.h>


// this structure describes the manager of events
typedef struct evd_ctrl_t evd_ctrl_t;

struct evd_ctrl_t{
    unsigned int SocksCount;
    // count of sockets that are actually checked by poll
    unsigned int chk_count;
    struct pollfd * Socks;
    size_t SocksSz;
    evd_t* *Cfgs;
    llist_t ready;
};

static evd_ctrl_t Evd;

static void
initiate_underlying_system(void)
{
    memset(&Evd, 0, sizeof(Evd));
    llist_init(&Evd.ready);
    Evd.SocksSz = 100;
    Evd.Socks = new_space(sizeof(*Evd.Socks) * Evd.SocksSz);
    Evd.Cfgs = new_space(sizeof(*Evd.Cfgs) * Evd.SocksSz);
}

static void
shutdown_underlying_system(void)
{
    if(Evd.Socks) free(Evd.Socks);
    if(Evd.Cfgs) free(Evd.Cfgs);
    memset(&Evd, 0, sizeof(Evd));
}

int
evtid_in_use(void)
{
    return Evd.Socks != NULL;
}

static void
start_checking_slot(unsigned idx)
{
    if(idx > Evd.chk_count)
    {
        evd_t * evd = Evd.Cfgs[idx];
        struct pollfd copy;

        copy = Evd.Socks[Evd.chk_count];
        Evd.Socks[Evd.chk_count] = Evd.Socks[idx];
        Evd.Socks[idx] = copy;

        Evd.Cfgs[idx] = Evd.Cfgs[Evd.chk_count];
        Evd.Cfgs[idx]->_data.index = idx;
        Evd.Cfgs[Evd.chk_count] = evd;
        evd->_data.index = Evd.chk_count;
    }
    Evd.chk_count++;
}

static void
stop_checking_slot(unsigned idx)
{
    Evd.Socks[idx].events = 0;
    Evd.chk_count--;
    // if the current element is not at the end, then I switch
    // 2 elements.
    if(idx != Evd.chk_count)
    {
        evd_t * evd = Evd.Cfgs[idx];
        struct pollfd copy;

        copy = Evd.Socks[Evd.chk_count];
        Evd.Socks[Evd.chk_count] = Evd.Socks[idx];
        Evd.Socks[idx] = copy;

        Evd.Cfgs[idx] = Evd.Cfgs[Evd.chk_count];
        Evd.Cfgs[idx]->_data.index = idx;
        Evd.Cfgs[Evd.chk_count] = evd;
        evd->_data.index = Evd.chk_count;
    }
}

static void 
call_real_manager(int wait_time)
{
    int chk;

    // we must have either checker or wait time. Otherwise
    // we'll end up in poll forever (or very long time)!
    assert(Evd.chk_count != 0 || wait_time >= 0);

    chk = poll(Evd.Socks, Evd.chk_count, wait_time);
    if(chk < 0)
    {
        if(errno != EINTR) error(EXIT_FAILURE, errno, "Unexpected error from poll");
    }
    else if(chk > 0)
    {
        unsigned int i;
        for(i = Evd.chk_count; i != 0 ; i--)
        {
            register struct pollfd * pfd = Evd.Socks + i - 1;
            evd_t * evd;
            evd_internal_t * slot;

            evd = Evd.Cfgs[i - 1];
            slot = &evd->_data;

            if( pfd->revents & (POLLIN|POLLHUP|POLLERR) )
            {
                SET_READ_READY(slot); 
                if(llist_empty(&slot->l)) llist_add(&Evd.ready, &slot->l);
                if(CHKS_READ(slot)) CLR_READ_CHK(slot);
            }
            if( pfd->revents & (POLLOUT|POLLHUP|POLLERR) )
            {
                SET_WRITE_READY(slot); 
                if(llist_empty(&slot->l)) llist_add(&Evd.ready, &slot->l);
                if(CHKS_WRITE(slot)) CLR_WRITE_CHK(slot);
            }
            // if this socket does not need to check anything else, then
            // I kick it out from the list of to be checked.
            if(!CHKS_WRITE(slot) && !CHKS_READ(slot))
                stop_checking_slot(i - 1);
        }
    }
}

void
evd_watch(evd_t * evd)
{
    register struct pollfd * pfd;
    evd_internal_t * slot = &evd->_data;

    assert(!IS_WATCHED(slot) && evd->d >= 0);

    SET_WATCHED(slot);

    // if there are no empty slots then shift SocksCount to the right
    if(Evd.SocksCount == Evd.SocksSz)
    {
	size_t new_sz = Evd.SocksSz + 100;
        Evd.Socks = more_space(Evd.Socks, new_sz * sizeof(*Evd.Socks));
	Evd.Cfgs = more_space(Evd.Cfgs, new_sz * sizeof(*Evd.Cfgs));
	Evd.SocksSz = new_sz;
    }
    pfd = Evd.Socks + Evd.SocksCount;
    pfd->fd = evd->d;
    pfd->events = 0;
    pfd->revents = 0;
    Evd.Cfgs[Evd.SocksCount] = evd;
    slot->index = Evd.SocksCount;
    Evd.SocksCount++;
}

void 
evd_nowatch(evd_t * evd)
{
    evd_internal_t * slot = &evd->_data;

    assert(IS_WATCHED(slot));

    if(slot->index < Evd.chk_count)
        stop_checking_slot(slot->index);

    if(!llist_empty(&slot->l)) llist_drop(&slot->l);

    Evd.SocksCount--;
    if(slot->index != Evd.SocksCount)
    {
	Evd.Socks[slot->index] = Evd.Socks[Evd.SocksCount];
	Evd.Cfgs[slot->index] = Evd.Cfgs[Evd.SocksCount];
	Evd.Cfgs[slot->index]->_data.index = slot->index;
    }

    slot->flags = 0;
}


// this function requests that read-readinnes is checked
void
evd_input_exhausted(evd_t * evd)
{
    register evd_internal_t * slot = &evd->_data;
    struct pollfd * pfd;

    assert(IS_WATCHED(slot));

    if(CHKS_READ(slot)) return;

    // remove read readinnes indication
    CLR_READ_READY(slot);
    SET_READ_CHK(slot);

    pfd = Evd.Socks + slot->index;
    pfd->events |= POLLIN;
    if(slot->index >= Evd.chk_count) start_checking_slot(slot->index);
}


// this function requests that write-readinnes is checked
void 
evd_check_waction(evd_t * evd)
{
    register evd_internal_t * slot = &evd->_data;
    struct pollfd * pfd;

    assert(IS_WATCHED(slot));

    // if the check is requested already, then we don't do anything else
    if(CHKS_WRITE(slot)) return;

    // just in case, clear write-readinness flag.
    CLR_WRITE_READY(slot);

    pfd = Evd.Socks + slot->index;
    pfd->events |= POLLOUT;
    if(slot->index >= Evd.chk_count) start_checking_slot(slot->index);

    SET_WRITE_CHK(slot);
}

