#include "evtid.h"
#include "xalloc.h"
#include <string.h>
#include <stddef.h>
#include <sys/select.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <assert.h>
#include <error.h>


// this structure describes the manager of events
typedef struct evd_ctrl_t evd_ctrl_t;

struct evd_ctrl_t{
    unsigned int SocksCount;
    size_t SocksSz;
    evd_t* *Socks;
    // count of sockets that are actually checked by select
    unsigned int chk_count;
    llist_t ready;
};

static evd_ctrl_t Evd;

static void
initiate_underlying_system(void)
{
    memset(&Evd, 0, sizeof(Evd));
    llist_init(&Evd.ready);
    Evd.SocksSz = 100;
    Evd.Socks = new_space(sizeof(*Evd.Socks) * Evd.SocksSz);
}

static void
shutdown_underlying_system(void)
{
    if(Evd.Socks) free(Evd.Socks);
    memset(&Evd, 0, sizeof(Evd));
}

int
evtid_in_use(void)
{
    return Evd.Socks != NULL;
}

static void
start_checking_slot(unsigned idx)
{
    if(idx > Evd.chk_count)
    {
        evd_t * evd = Evd.Socks[idx];

        Evd.Socks[idx] = Evd.Socks[Evd.chk_count];
        Evd.Socks[idx]->_data.index = idx;
        Evd.Socks[Evd.chk_count] = evd;
        evd->_data.index = Evd.chk_count;
    }
    Evd.chk_count++;
}

static void
stop_checking_slot(unsigned idx)
{
    Evd.chk_count--;
    // if the current element is not at the end, then I switch
    // 2 elements.
    if(idx != Evd.chk_count)
    {
        evd_t * evd = Evd.Socks[idx];

        Evd.Socks[idx] = Evd.Socks[Evd.chk_count];
        Evd.Socks[idx]->_data.index = idx;
        Evd.Socks[Evd.chk_count] = evd;
        evd->_data.index = Evd.chk_count;
    }
}

static void 
call_real_manager(int wait_time)
{
    int chk;
    struct timeval tout, * tout_ptr;
    fd_set rfds, wfds, efds;
    unsigned i;
    int nfds;
    

    if(wait_time >= 0)
    {
        tout.tv_sec = wait_time / 1000;
        tout.tv_usec = (wait_time % 1000) * 1000;
        tout_ptr = &tout;
    }
    else
    {
        tout_ptr = NULL;
        assert(Evd.chk_count != 0);
    }

    FD_ZERO(&rfds);
    FD_ZERO(&wfds);
    FD_ZERO(&efds);

    nfds = 0;
    for(i = 0; i < Evd.chk_count; i++)
    {
        evd_t * evd = Evd.Socks[i];
        evd_internal_t * slot = &evd->_data;
        if(CHKS_READ(slot))
        {
            FD_SET(evd->d, &rfds);
            FD_SET(evd->d, &efds);
        }
        if(CHKS_WRITE(slot))
        {
            FD_SET(evd->d, &wfds);
            FD_SET(evd->d, &efds);
        }
        if(evd->d >= nfds) nfds = evd->d + 1;
    }

    chk = select(nfds, &rfds, &wfds, &efds, tout_ptr);
    if(chk < 0)
    {
        if(errno != EINTR) error(EXIT_FAILURE, errno, "Unexpected error from select");
    }
    else if(chk > 0)
    {
        for(i = Evd.chk_count; i != 0 ; i--)
        {
            evd_t * evd = Evd.Socks[i-1];
            evd_internal_t * slot = &evd->_data;

            if( FD_ISSET(evd->d, &rfds) || FD_ISSET(evd->d, &efds))
            {
                SET_READ_READY(slot); 
                if(llist_empty(&slot->l)) llist_add(&Evd.ready, &slot->l);
                if(CHKS_READ(slot)) CLR_READ_CHK(slot);
            }
            if( FD_ISSET(evd->d, &wfds) || FD_ISSET(evd->d, &efds))
            {
                SET_WRITE_READY(slot); 
                if(llist_empty(&slot->l)) llist_add(&Evd.ready, &slot->l);
                if(CHKS_WRITE(slot)) CLR_WRITE_CHK(slot);
            }
            // if this socket does not need to check anything else, then
            // I kick it out from the list of to be checked.
            if(!CHKS_WRITE(slot) && !CHKS_READ(slot))
                stop_checking_slot(i - 1);
        }
    }
}

void
evd_watch(evd_t * evd)
{
    evd_internal_t * slot = &evd->_data;

    assert(!IS_WATCHED(slot) && evd->d >= 0);

    SET_WATCHED(slot);

    // if there are no empty slots then shift SocksCount to the right
    if(Evd.SocksCount == Evd.SocksSz)
    {
	size_t new_sz = Evd.SocksSz + 100;
        Evd.Socks = more_space(Evd.Socks, new_sz * sizeof(*Evd.Socks));
	Evd.SocksSz = new_sz;
    }
    Evd.Socks[Evd.SocksCount] = evd;
    slot->index = Evd.SocksCount;
    Evd.SocksCount++;
}

void 
evd_nowatch(evd_t * evd)
{
    evd_internal_t * slot = &evd->_data;

    assert(IS_WATCHED(slot));

    if(slot->index < Evd.chk_count)
        stop_checking_slot(slot->index);

    if(!llist_empty(&slot->l)) llist_drop(&slot->l);

    Evd.SocksCount--;
    if(slot->index != Evd.SocksCount)
    {
	Evd.Socks[slot->index] = Evd.Socks[Evd.SocksCount];
	Evd.Socks[slot->index]->_data.index = slot->index;
    }

    slot->flags = 0;
}


// this function requests that read-readinnes is checked
void
evd_input_exhausted(evd_t * evd)
{
    register evd_internal_t * slot = &evd->_data;

    assert(IS_WATCHED(slot));

    if(CHKS_READ(slot)) return;

    // remove read readinnes indication
    CLR_READ_READY(slot);

    SET_READ_CHK(slot);
    if(slot->index >= Evd.chk_count) start_checking_slot(slot->index);
}


// this function requests that write-readinnes is checked
void 
evd_check_waction(evd_t * evd)
{
    register evd_internal_t * slot = &evd->_data;

    assert(IS_WATCHED(slot));

    // if the check is requested already, then we don't do anything else
    if(CHKS_WRITE(slot)) return;

    // just in case, clear write-readinness flag.
    CLR_WRITE_READY(slot);

    SET_WRITE_CHK(slot);
    if(slot->index >= Evd.chk_count) start_checking_slot(slot->index);
}

