#include "evtid.h"
#include "predok.h"
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <error.h>
#include <assert.h>
#include <unistd.h>
#include <fcntl.h>

#ifdef EVTI_WITH_ITIMER
#include <signal.h>
#endif

#ifndef TMR_GROUPS_CNT
#define TMR_GROUPS_CNT 64
#endif
#ifndef TMR_GROUP_SIZE
#define TMR_GROUP_SIZE 64
#endif

#define TMR_LOOP ((TMR_GROUPS_CNT) * (TMR_GROUP_SIZE))

// this structure describes the manager of events
typedef struct evti_ctrl_t evti_ctrl_t;


struct evti_ctrl_t{
    llist_t Group[TMR_GROUPS_CNT];
    llist_t Single[TMR_GROUP_SIZE];
    size_t count; // total count of registered timers
    size_t Group_idx; // index into Group array
    size_t Single_idx; // index into Single array
    unsigned delay; // delay in ms that we have to sleep. It shall never be longer than TMR_LOOP
    // this structure shall contain value of itimer from the last check
    evti_mark_t StartMark;
    uint64_t LastDelay;
#ifdef EVTI_WITH_ITIMER
    sig_atomic_t wraps;
#endif
};

static evti_ctrl_t Host;

#ifndef DEBUG_TIMER
#define tracer(tmr, type, off) (void) 0
#define report_value(...) (void) 0
#else

static inline void
tracer(evti_internal_t * meta, const char * type, int off)
{
    struct timeval n;
    evti_t * tmr = predok(meta, evti_t, _data);
    gettimeofday(&n, NULL);
    fprintf(stderr, "%ld.%06ld, %d:%d, %p with delay %d goes into %s + %d\n",
	    n.tv_sec, n.tv_usec, (int)Host.Single_idx, (int)Host.Group_idx,
	    tmr, (int)meta->period, type, off);
}

#define report_value(msg, ...) do{\
    struct timeval n;\
    gettimeofday(&n, NULL);\
    fprintf(stderr, "%ld.%06ld " msg "\n", n.tv_sec, n.tv_usec, ##__VA_ARGS__);\
}while(0)

#endif


#ifndef EVTI_WITH_ITIMER

uint64_t
evti_delay(const evti_mark_t * mark)
{
    struct timespec now;
    uint64_t delay;

    if(clock_gettime(CLOCK_MONOTONIC, &now) != 0)
	error(EXIT_FAILURE, errno, "Can't obtain CLOCK_MONOTONIC");

#ifdef DEBUG_TIMER
    {
	struct timeval n;
	gettimeofday(&n, NULL);
    fprintf(stderr, "%ld.%06ld evti_delay mark %ld.%09ld\n", n.tv_sec, n.tv_usec, mark->tv_sec, mark->tv_nsec);
    fprintf(stderr, "%ld.%06ld evti_delay now  %ld.%09ld\n", n.tv_sec, n.tv_usec, now.tv_sec, now.tv_nsec);
    }
#endif

    assert(mark->tv_sec < now.tv_sec || (mark->tv_sec == now.tv_sec && mark->tv_nsec <= now.tv_nsec));

    if(now.tv_nsec < mark->tv_nsec)
    {
	now.tv_sec--;
	now.tv_nsec += 1000000000;
    }

    delay = (uint64_t)(now.tv_sec - mark->tv_sec) * 1000 + (now.tv_nsec - mark->tv_nsec + 500000) / 1000000;

#ifdef DEBUG_TIMER
    fprintf(stderr, "evti_delay is %llu\n", (unsigned long long)delay);
#endif
    return delay;
}

void
evti_advance_mark(evti_mark_t * mark, uint64_t msec)
{
    if(msec == 0) return;

    mark->tv_sec += msec / 1000;
    msec %= 1000;
    // since when we calculate the delay we sometimes add 500, here we shall
    // substruct 500 usec.
    if(msec == 0)
    {
        mark->tv_sec--;
        mark->tv_nsec = 1000000000 - 500000;
    }
    else
    {
        mark->tv_nsec += (msec % 1000) * 1000000 - 500000;
        if(mark->tv_nsec > 999999999)
        {
            mark->tv_sec++;
            mark->tv_nsec -= 1000000000;
        }
    }
}

void
evti_delay_mark(evti_mark_t * mark)
{
    if(clock_gettime(CLOCK_MONOTONIC, mark) != 0)
	error(EXIT_FAILURE, errno, "Can't obtain CLOCK_MONOTONIC");
}

#else

#ifndef ITM_TV_SEC
#define ITM_TV_SEC 64
#endif

// Since we don't want our wraps counter to wrap around, we
// should get SIGALRM at least once in 60 seconds. This shall allow
// many years of running without wrapping :)
#if ITM_TV_SEC < 60
#error "The value for ITM_TV_SEC should not be less than 60"
#endif

static uint64_t
time_diff(const struct timeval *l, const struct timeval *r)
{
    register uint64_t sec = l->tv_sec - r->tv_sec;
    register long usec;
    if(l->tv_usec < r->tv_usec)
    {
	usec = 1000000 + l->tv_usec - r->tv_usec;
	sec--;
    }
    else
	usec = l->tv_usec - r->tv_usec;
    return sec * 1000 + (usec + 500) / 1000;
}


uint64_t
evti_delay(const evti_mark_t * mark)
{
    struct itimerval now;
    register struct timeval * np;
    register const struct timeval *rec;
    uint64_t delay = 0;

    getitimer(ITIMER_REAL, &now);
    np = &now.it_value;
    rec = &mark->tm;

#ifdef DEBUG_TIMER
    fprintf(stderr, "evti_delay mark .wrap %d, at %ld.%06ld\n", mark->wrap, rec->tv_sec, rec->tv_usec);
    fprintf(stderr, "evti_delay now  .wrap %d, at %ld.%06ld\n", Host.wraps, np->tv_sec, np->tv_usec);
#endif

    assert(mark->wrap <= Host.wraps);

    if(mark->wrap < (int)Host.wraps)
	delay += (uint64_t)(Host.wraps - mark->wrap) * ITM_TV_SEC * 1000;


    if(rec->tv_sec < np->tv_sec
	    || (rec->tv_sec == np->tv_sec && rec->tv_usec < np->tv_usec))
    {
        uint64_t diff;

        // if the itimer had to be restarted after the fork, then the now
        // moment might be earlier than the recorded moment. In this case
        // we return 0.
        if(delay == 0) return 0;

	diff = time_diff(np, rec);
        assert(diff <= delay);
	delay -= diff;
    }
    else delay += time_diff(rec, np);

#ifdef DEBUG_TIMER
    fprintf(stderr, "evti_delay is %llu\n", (unsigned long long)delay);
#endif
    return delay;
}

void
evti_advance_mark(evti_mark_t * mark, uint64_t msec)
{
    register struct timeval *rec;
    long sec, usec;

    if(msec == 0) return;

    rec = &mark->tm;
    sec = msec / 1000;
    usec = (msec % 1000) * 1000;
    // since when we calculate the delay we add 500, here we shall
    // substruct 500 usec.
    if(usec == 0)
    {
        assert(sec != 0);
        sec--;
        usec = 1000000 - 500;
    }
    else usec -= 500;
    
    if(sec >= ITM_TV_SEC)
    {
	mark->wrap += sec / ITM_TV_SEC;
	sec %= ITM_TV_SEC;
    }

    if(usec > rec->tv_usec)
    {
	rec->tv_usec += 1000000;
	if(rec->tv_sec == 0)
	{
	    rec->tv_sec = ITM_TV_SEC;
	    mark->wrap++;
	}
	rec->tv_sec--;
    }
    rec->tv_usec -= usec;
    if(sec > rec->tv_sec)
    {
	rec->tv_sec += ITM_TV_SEC;
	mark->wrap++;
    }
    rec->tv_sec -= sec;
}

void
evti_delay_mark(evti_mark_t * mark)
{
    struct itimerval now;

    getitimer(ITIMER_REAL, &now);
    mark->tm = now.it_value;
    mark->wrap = (int)Host.wraps;
}


#endif



static void
advance_group()
{
    llist_t * el;
    llist_t * base = Host.Group + Host.Group_idx;

    el = llist_left(base);
    while(el != base)
    {
	evti_internal_t * meta = predok(el, evti_internal_t, l);

	el = llist_left(el);

	if(meta->period >= TMR_LOOP)
	{
	    meta->period -= TMR_LOOP;
	    tracer(meta, "(dec loop) group", (int)Host.Group_idx);
	}
	else
	{
	    register llist_t * slot;
	    llist_drop(&meta->l);
	    tracer(meta, "(advance) seconds", (int)meta->period);
	    slot = Host.Single + (size_t)meta->period;
	    llist_add(Host.Single + (size_t)meta->period, &meta->l);
	}
    }
    Host.Group_idx++;
    if(Host.Group_idx == TMR_GROUPS_CNT) Host.Group_idx = 0;
}


// this function shall quickly skip all the timer slots that don't 
// have timers. It must never have a need to do advance_group or rehash_all
static void
quick_advance(uint64_t elapsed)
{
    if(elapsed == 0) return;
    report_value("quick_advance elapsed %llu", (unsigned long long)elapsed);
    assert(elapsed < TMR_LOOP);

    elapsed += Host.Single_idx + Host.Group_idx * TMR_GROUP_SIZE;
    if(elapsed >= TMR_LOOP) elapsed -= TMR_LOOP;

    Host.Single_idx = elapsed % TMR_GROUP_SIZE;
    Host.Group_idx = elapsed / TMR_GROUP_SIZE;
    report_value("After quick_advance: Single_idx %llu, Group_idx %llu", 
	    (unsigned long long)Host.Single_idx, (unsigned long long)Host.Group_idx);
}


static inline void
set_new_delay(void)
{
    size_t j;
    Host.delay = 0;
    // start with current slot and go to the end, looking for non-empty slots
    for(j = Host.Single_idx; j < TMR_GROUP_SIZE && llist_empty(Host.Single + j); j++) Host.delay++;

    // if we didn't reach the end of single slots, or the next group
    // slot is non-empty, then we can leave
    if(j != TMR_GROUP_SIZE || !llist_empty(Host.Group + Host.Group_idx)) 
	return;


    // ok. we can start checking from the beginning of the single groups
    for(j = 0; j < Host.Single_idx && llist_empty(Host.Single + j); j++) Host.delay++;


    // if we didn't make full loop, then it means we found non-empty slot, so
    // we are done
    if(j != Host.Single_idx) return;

    // at this point we know that all Single slots are empty.
    // The next group is empty, so we know that we can sleep TMR_GROUP_SIZE
    // plus whatever is left in the current group
    Host.delay = 2 * TMR_GROUP_SIZE - Host.Single_idx;

    // don't check the next group, we know it is empty. 
    for(j = Host.Group_idx + 1; j < TMR_GROUPS_CNT && llist_empty(Host.Group + j); j++)
	Host.delay += TMR_GROUP_SIZE;

    // again if we have found non-empty slot we are done
    if(j != TMR_GROUPS_CNT) return;

    // otherwise we check our group slots from start
    for(j = 0; j < Host.Group_idx && llist_empty(Host.Group + j); j++) 
	Host.delay += TMR_GROUP_SIZE;

    // finally if nothing is found then we simply set our delay to TMR_LOOP
    // since normally it shall never exceed it
    if(j == Host.Group_idx) Host.delay = TMR_LOOP;
}


static inline void
switch_lists(llist_t * new, llist_t * old)
{
    llist_add(new, old);
    llist_drop(old);
}

// this function shall be called after poll has returned. It shall check which 
// timers have expired. Actually it shall do noting if it is still below
// watermark.
static void
tmrs_handler(void)
{
    uint64_t abs_elapsed, elapsed;

    // if there are no timers, then we do nothing
    if(Host.count == 0) return;

    abs_elapsed = evti_delay(&Host.StartMark);

    // looks like from time to time we get abs_elapsed less than LastDelay.
    // Though so far it was seen only on 32 bit systems.
    if(abs_elapsed >= Host.LastDelay)
    {
        elapsed = abs_elapsed - Host.LastDelay;
        Host.LastDelay = abs_elapsed;
    }
    else elapsed = 0;

    report_value("in tmrs_handler elapsed %llu, sid %zu, gid %zu", 
	    (unsigned long long)elapsed, Host.Single_idx, Host.Group_idx);
    // if the desired delay didn't expire yet then we simply advance our
    // indexes
    if(elapsed < Host.delay) 
    {
	Host.delay -= elapsed;
	quick_advance(elapsed);
    }
    else
    {
	llist_t tmrs;

	llist_init(&tmrs);
	
	if(Host.delay == 0)
	{
	    // well, the timer could have been droped, so we check
	    if(!llist_empty(Host.Single + Host.Single_idx))
		switch_lists(&tmrs, Host.Single + Host.Single_idx);
	}
	else if(Host.delay > 1)
	{
	    assert(llist_empty(Host.Single + Host.Single_idx));
	    elapsed -= Host.delay - 1;
	    // quickly skip all empty slots except for the last one.
	    // this one shall be just before the non-empty, so we have
	    // to do the directory advancing and so on
	    quick_advance(Host.delay - 1);
	}

	while(elapsed)
	{
	    Host.Single_idx++;
	    elapsed--;
	    if(Host.Single_idx == TMR_GROUP_SIZE)
	    {
		Host.Single_idx = 0;
		advance_group();
	    }
	    if(!llist_empty(Host.Single + Host.Single_idx))
		switch_lists(&tmrs, Host.Single + Host.Single_idx);
	}

	set_new_delay();

	// now we can execute all expired timers
	while(!llist_empty(&tmrs))
	{
	    register evti_t * tmr = predok(llist_left(&tmrs), evti_t, _data.l);
	    evti_stop(tmr);
	    tmr->action(tmr);
	} 
    }
}


// this function figures out how long poll shall sleep.
static int
wait_time_val(void)
{
    int rv = 0;

    // if there are no timers then negative is returned
    if(Host.count == 0) rv = -1;

    // otherwise we check the value only if it is non-zero
    else if(Host.delay != 0)
    {
        // we check how much time elapsed since the processing of timers was
        // started
        uint64_t elapsed = evti_delay(&Host.StartMark) - Host.LastDelay;
        // if the time is higher than delay, then we return 0, otherwise
        // we just substract elapsed time from the desired one and return that
        // value.
        if(elapsed < (uint64_t) Host.delay) rv = Host.delay - (unsigned)elapsed;
    }
    report_value("wait time %d", rv);
    return rv;
}

#ifdef EVTI_WITH_ITIMER
static void
add_wrap(int no)
{
    Host.wraps++;
}

static void
evti_start_itimer(void)
{
    struct sigaction si;
    struct itimerval val = {{0}};

    sigfillset(&si.sa_mask);
    si.sa_flags = SA_RESTART;
    si.sa_handler = add_wrap;
    if( sigaction(SIGALRM, &si, NULL) != 0) 
        error(EXIT_FAILURE, errno, "Can't set signal action");
    val.it_value.tv_sec = ITM_TV_SEC;
    val.it_value.tv_usec = 0;
    val.it_interval.tv_sec = ITM_TV_SEC;
    val.it_interval.tv_usec = 0;
    fprintf(stderr, "Starting itimer\n");
    if( setitimer(ITIMER_REAL, &val, NULL) != 0)
        error(EXIT_FAILURE, errno, "Can't start itimer");
}
#endif

static void
tmrs_ctrl_init(void)
{
    size_t i;

    memset(&Host, 0, sizeof(Host));

    for(i = 0; i < TMR_GROUPS_CNT; i++) llist_init(Host.Group + i);
    for(i = 0; i < TMR_GROUP_SIZE; i++) llist_init(Host.Single + i);
}

void 
evti_start(evti_t * tmr, uint64_t period)
{
    llist_t * slot;
    evti_internal_t * meta = &tmr->_data;

    assert(!evti_running(tmr));

    report_value("start timer %p for %llu usec", tmr, (unsigned long long)period);

    // if this is the very first timer to add, then we reset our system
    if(Host.count == 0)
    {
	evti_delay_mark(&Host.StartMark);
	Host.LastDelay = 0;
	Host.Single_idx = 0;
	Host.Group_idx = 0;
	Host.delay = TMR_LOOP;
    }

    if(period < TMR_GROUP_SIZE)
    {
	size_t sid;
	sid = (Host.Single_idx + period) % TMR_GROUP_SIZE;
	tracer(meta, "seconds", (int)sid);
	slot = Host.Single + sid;
	if(period < Host.delay) Host.delay = period;
    }
    else
    {
	uint64_t tmp;
	size_t gid;
        unsigned diff;

        // substruct from the period the amount of time that is left till the
        // current group will be rotated.
        tmp = period - TMR_GROUP_SIZE + Host.Single_idx;

        // mark the number of full loops that the system has to make for this
        // period.
        meta->period = (tmp / TMR_LOOP) * TMR_LOOP;

        // now take the rest of it
        tmp %= TMR_LOOP;

        // this shall be the count of groups for the period
        gid = tmp / TMR_GROUP_SIZE;

        // mark the number of seconds that exceed the group
        meta->period += tmp % TMR_GROUP_SIZE;

        // new delay is simple. Just take whatever is left in the current
        // group and add number of groups multiplied by group size
        diff = TMR_GROUP_SIZE - Host.Single_idx + gid * TMR_GROUP_SIZE;
	if(Host.delay > diff) Host.delay = diff;

        // to find the slot, add the number of groups to the current group
        // and don't forget to rotate it
        gid += Host.Group_idx;
        gid %= TMR_GROUPS_CNT;
        slot = Host.Group + gid;

	tracer(meta, "group", (int)gid);
    }
    report_value("Host.delay %u", Host.delay);
    llist_add(slot, &meta->l);
    Host.count++;
}



#ifdef EVTI_WITH_ITIMER
static void
evti_ctrl_fini(void)
{
    struct sigaction si;
    struct itimerval val = {{0}};

    sigfillset(&si.sa_mask);
    si.sa_flags = SA_RESTART;
    si.sa_handler = SIG_DFL;
    sigaction(SIGALRM, &si, NULL);
    if(setitimer(ITIMER_REAL, &val, NULL) != 0)
        error(EXIT_FAILURE, errno, "Can't clear itimer");
}
#endif

void 
evti_stop(evti_t * tmr)
{
    evti_internal_t * meta = &tmr->_data;
    assert(evti_running(tmr));
    tracer(meta, "removed", 0);
    llist_drop(&meta->l);
    Host.count--;
}

/*
 * this part contains handling for descriptors and the events loop.
 * it depends on the underlying system, so I use some conditional includes
 */

#define SET_READ_CHK(inf) do{(inf)->flags |= 1;}while(0)
#define CHKS_READ(inf) ((inf)->flags & 1)
#define CLR_READ_CHK(inf) do{(inf)->flags &= ~1;}while(0)

#define SET_WRITE_CHK(inf) do{(inf)->flags |= 2;}while(0)
#define CHKS_WRITE(inf) ((inf)->flags & 2)
#define CLR_WRITE_CHK(inf) do{(inf)->flags &= ~2;}while(0)

#define SET_WATCHED(inf) do{(inf)->flags |= 4;}while(0)
#define IS_WATCHED(inf) ((inf)->flags & 4)
#define CLR_WATCHED(inf) do{(inf)->flags &= ~4;}while(0)

#define SET_READ_READY(inf) do{(inf)->flags |= 16; }while(0)
#define READ_READY(inf) ((inf)->flags & 16)
#define CLR_READ_READY(inf) do{(inf)->flags &= ~16;}while(0)

#define SET_WRITE_READY(inf) do{(inf)->flags |= 8; }while(0)
#define WRITE_READY(inf) ((inf)->flags & 8)
#define CLR_WRITE_READY(inf) do{(inf)->flags &= ~8;}while(0)

#define SET_READ_REPORT(inf) do{(inf)->flags |= 32;}while(0)
#define REPORTS_READ(inf) ((inf)->flags & 32)
#define CLR_READ_REPORT(inf) do{(inf)->flags &= ~32;}while(0)


#if defined EVD_WITH_EPOLL
#include "evd_epoll_src.h"
#elif defined EVD_WITH_SELECT
#include "evd_select_src.h"
#else
#include "evd_poll_src.h"
#endif

// this function allows read-readinnes reporting. I always assume that the socket is ready
// to be read.
void
evd_raction_active(evd_t * evd)
{
    register evd_internal_t * slot = &evd->_data;
    assert(IS_WATCHED(slot));
    SET_READ_REPORT(slot);
    if(!CHKS_READ(slot))
    {
        SET_READ_READY(slot);
        if(llist_empty(&slot->l)) llist_add(&Evd.ready, &slot->l);
    }
}

// this function forbids reporting of read-readinnes. I don't clear the READ_READY flag. There
// is no need to do it. Also I don't remove it from the ready list, it might be needed for write
// readinnes.
void
evd_raction_inactive(evd_t * evd)
{
    register evd_internal_t * slot = &evd->_data;
    assert(IS_WATCHED(slot));
    CLR_READ_REPORT(slot);
}


// this function requests, that the write-readinnes is faked
void
evd_call_waction(evd_t * evd)
{
    register evd_internal_t * slot = &evd->_data;

    assert(IS_WATCHED(slot));
    // if the check is requested already, then we don't do anything else
    if(CHKS_WRITE(slot)) return;
    SET_WRITE_READY(slot);
    if(llist_empty(&slot->l)) llist_add(&Evd.ready, &slot->l);
}

void 
evtid_dispatcher(void)
{
    int wait_time;
    llist_t lock;
    llist_t ready_list;

#ifdef EVTI_WITH_ITIMER
    // the itimer does not survive the fork, so it has to be started here.
    evti_start_itimer();
#endif

    llist_init(&lock);
    llist_init(&ready_list);

    // if we have some sockets in the ready list already, then we
    // shouldn't wait
    if(!llist_empty(&Evd.ready)) wait_time = 0;
    else wait_time = wait_time_val();

    while(Evd.SocksCount || wait_time >= 0)
    {
        call_real_manager(wait_time);

	tmrs_handler();

        llist_add(&ready_list, &Evd.ready);
        llist_drop(&Evd.ready);

	// the part below shall be done only when some sockets are ready.
	// otherwise we continue with the next loop iteration.
	while(!llist_empty(&ready_list))
	{
	    register evd_t * evd;
	    register evd_internal_t * slot;
            llist_t * el;

            el = llist_left(&ready_list);
            llist_drop(el);

            // add this element to the "lock" list. It shall be removed from this
            // list if the user has decided not to do anything with it anymore.
            // after the loop all of the ready elements shall be moved back into ready list.
            llist_add(&lock, el);

	    slot = predok(el, evd_internal_t, l);
            evd = predok(slot, evd_t, _data);

	    if( READ_READY(slot) && REPORTS_READ(slot)) 
	    {
		assert(evd->raction != NULL);
		evd->raction(evd);
                // check if the user has disabled the checking
                if(llist_empty(&lock)) continue;
	    }

	    if( WRITE_READY(slot) ) 
	    {
                CLR_WRITE_READY(slot);
		if(evd->waction != NULL)
                {
                    evd->waction(evd);
                    if(llist_empty(&lock)) continue;
                }
	    }
            
            llist_drop(el);

            // if this element still has READ_READY flag and REPORTS_READ
            // flag, then we put it in the ready list again. The WRITE_READY
            // may also be set.
            if((READ_READY(slot) && REPORTS_READ(slot)) || WRITE_READY(slot))
                llist_add(&Evd.ready, el);
	}
        if(!llist_empty(&Evd.ready)) wait_time = 0;
        else wait_time = wait_time_val();
    }
}

void
evd_prepare(int fd)
{
    fcntl(fd, F_SETFL, O_NONBLOCK);
    fcntl(fd, F_SETFD, FD_CLOEXEC);
}

void
evtid_init(void)
{
    initiate_underlying_system();
    tmrs_ctrl_init();
}

void
evtid_fini(void)
{
#ifdef EVTI_WITH_ITIMER
    evti_ctrl_fini();
#endif
    shutdown_underlying_system();
}
