/*
 * This is header file for manager of TImer and/or Descriptor EVents.
 * The API is defined in an assumption that there'll be only one event manager
 * in the process/thread. I can't really think of any case where more than
 * one should work :) So all internal structures are static. If you need it
 * differently, then you'll have to arrange thread specific storage, that's
 * fairly easy with modern compilers using __thread storage class keyword.
 *
 * None of the callbacks receive "void *" arguments. This is done on purpose.
 * Generally any callback knows about context that it should work in. Which
 * means that it knows where the appropriate ev(d|ti)_t structure is stored.
 * So, knowing it it's very easy to calculate the address of the structure
 * that contains the ev(d|ti)_t structure. The predok macro does
 * exactly that. So far I didn't encounter any program that would not satisfy
 * this rule.
 */
#ifndef EVTID_H
#define EVTID_H
#include "stddef.h"
#include <stdint.h>
#include <assert.h>
#include <sys/time.h>
#include <time.h>
#ifdef EVD_WITH_EPOLL
#include "evd_epoll.h"
#else // the select has the same header as poll!
#include "evd_poll.h"
#endif
#include "evti.h"

// this structure defines parameters of file descriptor events handling
typedef struct evd_t evd_t;
// this is type of function that shall be called when descriptor has some
// events
typedef void (*evd_handler_t)(evd_t*);

struct evd_t { 
    // this function shall be called when there's data available in buffer
    evd_handler_t raction;
    // this is called when the outgoing buffer is available
    evd_handler_t waction;
    // this is actual descriptor
    int d;
    // this is internally used data
    evd_internal_t _data;
};

// INFO
// this function initiates the structure. Must be called before
// evd_watch!
static inline void
evd_init(evd_t * evd)
{
    evd_internal_init(&evd->_data);
}

// INFO
// this function can be used to make the descriptor non-blocking and 
// making sure that it would be closed on exec call.
void evd_prepare(int);

// INFO
// this function deinitiates the structure. Must be called after
// evd_nowatch!
static inline void
evd_fini(evd_t * evd)
{
    evd_internal_fini(&evd->_data);
}


// INFO
// this function returns non-zero if the structure is watched for descriptor
// events. Zero otherwise. Implies that evd_init was called on the structure.
static inline int
evd_watched(evd_t * evd)
{
    return evd_internal_watched(&evd->_data);
}

// INFO
// this function initiates necessary structures for dispatching events
void evtid_init(void);

// INFO
// this function does actual event dispatching. It returns when there are no
// more timers and no sockets to be serviced.
void evtid_dispatcher(void);

// INFO
// this function deallocates any internal structures used by the events
// manager
void evtid_fini(void);

// INFO
// this function returns non-zero if the dispatcher is still running. It'll
// return 0 after the call to evtid_fini.
int evtid_in_use(void);

// INFO
// this function initiates event watching on a descriptor. All events shall
// be reported using provided control structure. 
// Actually the descriptor
// shall be watched only for errors. Any other watching shall be requested
// separately by evd_raction_active and evd_call_waction.
void evd_watch(evd_t * evd);


// INFO
// This function requests that read-readinnes is reported via raction callback.
// The descriptor is considered to be always ready for reading untill
// evd_input_exhausted is called.
void evd_raction_active(evd_t *);

// INFO
// this function disables reporting of read-readinnes
void evd_raction_inactive(evd_t *);

// INFO
// This function shall be called when the read has returned EAGAIN. It clears the
// readinnes flag from the structure. The flag shall be set again only after the data
// becomes available for reading
void evd_input_exhausted(evd_t *);


// INFO
// this function requests that write-readinnes is checked
// when the socket becomes write ready, then the waction handler
// is called. Note, the handler is called only once. No more checking shall be done
void evd_check_waction(evd_t *);

// INFO
// this function requests that the waction is called again. Note, if the
// evd_check_waction is called before or after this function, then the check
// for readinnes is done. Otherwise, the waction shall be called without any
// check.
void evd_call_waction(evd_t *);

// INFO
// this function stops watching for events on the descriptor
void evd_nowatch(evd_t *);

/* TIMER API */
// this structure defines parameters of timer
typedef struct evti_t evti_t;
// this is type of function that shall be called when timer expires
// the function gets as parameter pointer to the timer.
typedef void (*evti_handler_t)(evti_t *);

struct evti_t {
    // this is function that shall be called when timer expires
    evti_handler_t action;
    evti_internal_t _data;
};

// INFO 
// this function initiates specified timer structure. 
// Must be called before evti_start
static inline void 
evti_init(evti_t * tmr)
{
    evti_internal_init(&tmr->_data);
}

// INFO
// this function returns non-zero if the timer is in the middle of count-down.
// It returns zero otherwise
static inline int 
evti_running(evti_t * tmr)
{
    return evti_internal_status(&tmr->_data);
}

// INFO
// this function starts the timer. After the amount of time specified expires
// the handler shall be called. When the handler is called the timer is
// already stopped. In other words evti_running shall return 0.
// The period specifies number of ms to wait. Which means that the smallest
// non-zero amount to sleep is 1 ms. This might be changed with time.
void evti_start(evti_t * tmr, uint64_t period);
static inline void
evti_start_ms(evti_t * tmr, unsigned long ms)
{
    evti_start(tmr, (uint64_t)ms);
}

static inline void
evti_start_us(evti_t * tmr, unsigned long us)
{
    evti_start(tmr, (uint64_t)(us / 1000));
}

static inline void
evti_start_s(evti_t * tmr, unsigned long sec)
{
    uint64_t ms;
    ms = sec;
    ms *= 1000;
    evti_start(tmr, ms);
}

// INFO
// this function stops count-down for specified timer. The timer should have
// been started. Otherwise it will generate exception.
void evti_stop(evti_t * tmr);

#ifndef EVTI_WITH_ITIMER
typedef struct timespec evti_mark_t;
#else
typedef struct evti_mark_t evti_mark_t;
struct evti_mark_t{
    struct timeval tm;
    int wrap;
};
#endif

// INFO
// use this function to mark the moment of time
void evti_delay_mark(evti_mark_t *);

// INFO
// use this function to add some value to the mark
void evti_advance_mark(evti_mark_t * , uint64_t );

// INFO
// this function shall tell you how many milliseconds expired since the
// time was marked by evti_delay_mark. It automatically marks the time
// again, so there's no need to call evti_delay_mark again.
uint64_t evti_delay(const evti_mark_t *);

#endif
