#!/usr/bin/perl
# this script generates Makefiles (or files to be included in Makefile)
# from the set of MMaker.hints files located in different directories.
# The goal is to provide a tool that simplifies Makefile creation in
# the situation when the sources are located in different directories.
#
# The script reads the MMaker.hints in the current directory, or in the directory
# specified as the argument and writes to STDOUT the Makefile that can be used
# to create all of the targets.
#
# If there's some error, then the execution is aborted and
# the error is written into STDERR.
#
# This is description of the MMaker.hints format
#
# The file is regular perl code. The code is executed with working directory
# set to the one where the file is located.
#
# There are few ways to provide hints for making Makefile.

# 1. The list of sources shall be in predeclared array @sources. Without it, all 
# .c files in the directory are used. The list may contain other directories
# which shall be parsed and sources from them added to this list.

# 2. If the sources in this directory use headers from other directories, then
# you can give the list of those directories using @header_dirs array. Each of the
# directories shall be parsed as well, and if it also contains @header_dirs array,
# then the content of that array shall be added to the list of header directories.

# 3. the hash %CPPFLAGS may be used to set flags for compiling 
# of different files. The hash shall contain file name as key and the string
# with flags as value. The flags shall be added to all compilations
# where specified file participates. The file can be some header file which is
# used by sources, then the flags shall be used in compiling all sources that
# depend on this header.

# 4. the hash %LDLIBS may be used to set libraries for linking
# of different files. The hash shall contain file name as key and the text
# with list of libraries as value. The list shall be added to all linking
# where specified file or its deriative participates 

# 5. the hash %LDFLAGS may be used to set flags for linking
# of different files. The hash shall contain file name as key and the string
# with flags as value. The flags shall be added to all linking
# where specified file or its deriavative participates

# 6. The CompilingFlags function can be used to add some values to CPPFLAGS
# variable. These values are collected globally, then unique keys are set
# as original value for CPPFLAGS which is used in all compilations.

# 7. If the compilation of all sources in some directory needs certain CPPFLAGS,
# then these can be given in array @sources_flags.

# 8. To specify what shall be the target(s) for make execution the functions
# Program and SharedLib can be used. These functions take as arguments key-value
# pairs. Common keys are
#
# name => this is the name of executable or library that shall be created. Note,
# if the shared library name ends with .so.NUM then the link file with name
# ending in .so is also created by makefile.
# 
# objects => points to reference to array with the list of sources from which the
# object files shall be created. The list may contain directories, in which case
# the sources shall be obtained from those directories.
#
# ldflags => this is string with additional values for LDFLAGS.
# 
# 9. The function Header requests creation of header. Again, the key-value
# pairs specify parameters for creation. The key "name" specifies the path
# to the file for creation. The key "headers" shall point to the reference
# to array containing either file name, or the directories from which the
# header files shall be taken. If the directory contains MMaker.hints with
# @headers array, then files from that array is taken. Otherwise the list of
# all headers in that directory.

# any code can be used to decide which flags shall be used. This script
# provides function "runs_fine". This function attempts to compile, link and execute
# specified code. It returns 1 if all steps worked. Otherwise 0. The argument
# to the function is reference to hash. The key "code" contains the code to
# be processed, the key "ldlibs" defines the libraries for linking, the
# key "cppflags" defines the flags for compiling. The key ldflags defines the flags
# for linking.

use strict;
use Cwd;
use Carp;
use File::Temp;
use Dumpvalue;
use Carp;

our $HintsFile = "MMaker.hints";

if(@ARGV)
{
    chdir($ARGV[0]) || die "Can't go to '$ARGV[0]': $!\n";
}

# I shall keep information about parsed directories in this
# hash. This shall allow me to skip parsing of same directory multiple
# times. The key in the hash is relative to Makefile path.
our %Dirs;

# this hash is needed to make sure that created object files won't override each other.
# so, the key is the name of object file that shall be created in current directory. The
# value is reference to array with first element pointing to directory info, second
# element containing source file name (this is needed for figuring out if there are any flags
# associated with this name). The relative path to the file is stored in the third element.
# Remember, the sources for some directory may point to file in another directory!)
# the fourth element shall contain hash with all flags for compiling this file.
# the fifth element shall be the hash with all header files that it uses (relative path)
our %Objs;

# in these hashes I copy data from hashes defined in individual MMaker.hints. As key I
# use relative path to the file, as value collected flags. It is an error to give flags
# for the same source file in different directories
our %LdFlags;
our %CppFlags;
our %LdLibs;

my $info = new_info();
my @progs;
sub wrapped::Program
{
    my %args = @_;
    my $msg = check_prog_or_lib(\%args, 'objects');
    croak($msg) if $msg;
    push(@progs, \%args);
}

our %GenericFlags;
sub wrapped::CompilingFlags
{
    foreach my $flag (@_)
    {
        $GenericFlags{$flag}++;
    }
}

my @libs;

sub wrapped::SharedLib
{
    my %args = @_;
    my $msg = check_prog_or_lib(\%args, 'objects');
    croak($msg) if $msg;
    push @libs, \%args;
}

# the list of files to be created
my @hdr_files;

sub wrapped::Header
{
    my %args = @_;
    my $msg = check_prog_or_lib(\%args, 'headers');
    croak($msg) if $msg;
    push @hdr_files, \%args;
}

# this function shall load the information about current directory in all hashes
load_directory('.');

# I don't care about programs and libraries defined in other directories, so I disable
# loading of those here.
*wrapped::Program = sub{};
*wrapped::SharedLib = sub{};
*wrapped::Header = sub{};

foreach my $hf (@hdr_files)
{
    my $name = $hf->{name};
    if(-e $name)
    {
        select(STDERR);
        local( $| ) = 1;
        print "Header '$name' already exists. Overwrite? (Y/n): ";
        select(STDOUT);
        my $ans = <STDIN>;
        next if $ans =~ /^\s*n/i;
    }
    make_all_dirs($name);
    my @files = make_headers_list($hf->{headers});

    open(HDR, ">$name") || die "Can't write header $name: $!\n";
    my $cwd = getcwd();
    my $dst_dir = $name;
    $dst_dir =~ s/\/+[^\/]+$//;
    foreach my $f (@files)
    {
        print HDR '#include "', shift_path($dst_dir, $f, $cwd), "\"\n";
    }
    close(HDR);
}

my $make_body;
my @make_targets;
my @clean_targets;

foreach my $prog (@progs)
{
    my $target = $prog->{name};
    push(@make_targets, $target);
    push(@clean_targets, $target);

    # just in case create all directories for storing the target
    make_all_dirs($target);

    # the the list of objects used for building this target
    my @obj_list = make_list_of_objects($prog);

    my $cmd = "\$(LINK.o)";
    $cmd .= " $prog->{ldflags}" if(exists $prog->{ldflags});
    $cmd .= " \$(LDFLAGS) \$^ -o \$@";
    $cmd .= " \$(LDLIBS)";

    $make_body .= "$target: ";
    $make_body .= join(' ', @obj_list);
    $make_body .= "\n\t$cmd\n";
    $make_body .= add_linking_flags($target, \@obj_list);
}

foreach my $lib (@libs)
{
    my $target = $lib->{name};

    # just in case create all directories for storing the target
    make_all_dirs($target);
    my $soname = $target;
    $soname =~ s/^.*\/+//;

    if($target =~ /^(.*\.so)(?:\.\d+)+$/)
    {
        $make_body .= "$1: $target\n\tln -s -f $soname $1\n";
        push(@make_targets, $1);
        push(@clean_targets, $1, $target);
    }
    else
    {
        push(@make_targets, $target);
        push(@clean_targets, $target);
    }

    # now I should go thru the list of all entities which shall be used as object
    # files for this target.
    my @obj_list = make_list_of_objects($lib);;

    my $cmd = "\$(LINK.o) -shared -nostartfiles -Wl,-z,defs -Wl,-soname,$soname";
    $cmd .= " -Wl,--version-script=$lib->{script}" if(exists $lib->{script});
    $cmd .= " $lib->{ldflags}" if(exists $lib->{ldflags});
    $cmd .= " \$(LDFLAGS) \$^ -o \$@";
    $cmd .= " \$(LDLIBS)";

    $make_body .= "$target: ";
    $make_body .= join(' ', @obj_list);
    $make_body .= "\n\t$cmd\n";
    foreach my $f (@obj_list)
    {
        $make_body .= "$f: CPPFLAGS+=-fpic\n";
    }

    $make_body .= add_linking_flags($target, \@obj_list);
}

die "No Programm or SharedLib defined\n" unless @make_targets;

# now I can add the commands for creating each of the object files
foreach my $f (keys %Objs)
{
    # add CPPFLAGS if necessary
    $make_body .= add_compiling_flags($f);
    my $obj = $Objs{$f};
    # add all of the dependency headers
    my @hdrs = $obj->used_headers;
    if(@hdrs)
    {
        $make_body .= "$f: ";
        $make_body .= join(' ', @hdrs);
        $make_body .= "\n";
    }
    my $spath = $obj->source_path;
    my $suffix = ($spath =~ /\.(\w+)$/) ? $1 : 'c';
    $make_body .= "$f: $spath\n\t\$(COMPILE.$suffix) $spath\n";
}

my @f = keys %GenericFlags;
print "CPPFLAGS=@f\n" if(@f);
print ".PHONY: all clean\n";
print "all: @make_targets\n";
print $make_body;
print "clean:\n";
print "\t-rm -f @clean_targets\n";
print "\t-rm -f @{[keys %Objs]}\n";

my $current_headers_dir;

sub make_headers_list
{
    my $list = shift;
    my @hdrs;
    foreach my $el (@$list)
    {
        if(-f $el)
        {
            push @hdrs, $el;
        }
        elsif(-d $el)
        {
            $current_headers_dir = '.';
            push @hdrs, headers_from_dir(compress_path($el), {});
        }
        else
        {
            die "The element '$el' in the headers list is neither file nor directory\n";
        }
    }
    return @hdrs;
}

sub headers_from_dir
{
    my $dir = shift;
    my $idx = shift;
    die "In $current_headers_dir: directory '$dir' is present in headers chain multiple times\n"
        if exists $idx->{$dir};
    $idx->{$dir}++;
    load_directory($dir);
    my @rv;
    foreach my $el (@{$Dirs{$dir}{headers}})
    {
        $current_headers_dir = $dir;
        push @rv, (-d $el) ? objects_from_dir($el, $idx) : $el;
    }
    return @rv;
}

my $current_sources_dir;
sub make_list_of_objects
{
    my $ctrl = shift;
    my @obj_list;
    foreach my $el (@{$ctrl->{objects}})
    {
        if(-f $el)
        {
            push @obj_list, object_from_file($el, '.');
        }
        elsif(-d $el)
        {
            $current_sources_dir = '.';
            push @obj_list, objects_from_dir(compress_path($el), {});
        }
        else
        {
            die "The element '$el' in the objects list is neither file nor directory\n";
        }
    }
    # it might happen, that the "header_dirs" is present among arguments
    # for building this target. I should add these directories to the objects info
    if(exists $ctrl->{header_dirs})
    {
        foreach my $f (@obj_list)
        {
            my $obj = $Objs{$f};
            $obj->add_header_dir(@{$ctrl->{header_dirs}});
        }
    }
    return @obj_list;
}

sub objects_from_dir
{
    my $dir = shift;
    my $idx = shift;
    die "In $current_sources_dir: directory '$dir' is present in sources chain multiple times\n"
        if exists $idx->{$dir};
    $idx->{$dir}++;
    load_directory($dir);
    my @rv;
    foreach my $el (@{$Dirs{$dir}{sources}})
    {
        $current_sources_dir = $dir;
        push @rv, (-d $el) ? objects_from_dir($el, $idx) : object_from_file($el, $dir);
    }
    return @rv;
}

sub object_from_file
{
    my $f = shift;
    my $dir = shift;
    die "File $f does not exist or is not regular file\n" unless -f $f;
    die "File '$f' contains spaces in name\n" if $f =~ /\s/;
    # as the object name I'll try to use the name of the source file.
    my $obj_name = $f;
    $obj_name =~ s/^.*\/([^\/]+)$/$1/;
    $obj_name .= '.o' unless $obj_name =~ s/\.\w+$/.o/;
    if(exists $Objs{$obj_name})
    {
        # if the object with this name already exists, then I make sure
        # that it is created from the same source. If yes, then I reuse it.
        my $obj = $Objs{$obj_name};
        return $obj_name if($obj->source_path eq $f);

        # well, I have to modify object name so that it does not conflict
        my $pref = "aa";
        while(exists $Objs{"${pref}_$obj_name"}){ $pref++; }
        $obj_name = join("_", $pref, $obj_name);
    }
    my $obj = ObjHolder->new($f, $Dirs{$dir});
    $Objs{$obj_name} = $obj;

    # now I need to obtain the list of header files on which this source depends
    find_headers($obj, $dir);
    return $obj_name;
}

sub find_headers
{
    my $obj = shift;
    my $f = $obj->source_path;
    my $dir = shift;
    my %flags;
    $flags{$CppFlags{$f}} = 1 if(exists $CppFlags{$f});
    foreach my $f ($obj->compile_flags())
    {
        $flags{$f} = 1;
    }

    my $src_dir = ($f =~ /^(.*)\//) ? compress_path($1) : '.';
    my @header_dirs = expand_header_dirs($dir);

    headers_check:
    my @flags = keys %flags;
    my $dep = `gcc -MM -MG @flags $f`;
    my ($t, $list) = split(/:\s*/, $dep, 2);
    my @h = grep /\.h$/, split /\s+/, $list;
    my @missing = ();
    my $need_recheck = 0;
    foreach my $hdr_file (@h)
    {
        # the existing headers are always listed with directories.
        # so, if the $src_dir is not current one, and the header name does not
        # contain directory name, then this header is missing.
        # If the $src_dir is current one, then I have to check which files really exist
        if(index($hdr_file, "/") < 0 && ($src_dir ne '.' || !-e $hdr_file))
        {
            push(@missing, $hdr_file);
            next;
        }
        if(exists $CppFlags{$hdr_file} && !exists $flags{$CppFlags{$hdr_file}})
        {
            $flags{$CppFlags{$hdr_file}} = 1;
            $need_recheck = 1;
        }
    }
    # if some header files are missing, then I have to find them and repeat the 
    # check. For finding them I use the header_dirs list from this directory
    if(@missing)
    {
        foreach my $hdr_dir (@header_dirs)
        {
            my @not_found = grep {!-e "$hdr_dir/$_"} @missing;
            if($#not_found != $#missing)
            {
                $flags{"-I$hdr_dir"} = 1;
                $obj->add_header_dir($hdr_dir);
            }
            @missing = @not_found;
            goto headers_check if(!@missing);
        }
        die "$f: Can't find header files @missing\n";
    }
    goto headers_check if $need_recheck;
    $obj->set_headers(\@h);
}

sub expand_header_dirs
{
    my $dir = shift;
    my @list = @{$Dirs{$dir}{header_dirs}};
    my $idx = @_ ? shift : {};
    my @rv = @list;
    foreach my $d (@list)
    {
        next if exists $idx->{$d};
        $idx->{$d} = 1;
        load_directory($d);
        push @rv, expand_header_dirs($d, $idx);
    }
    return @rv;
}

sub add_compiling_flags
{
    my $f = shift;
    my $obj = $Objs{$f};
    my $spath = $obj->source_path;
    my @flags = $obj->compile_flags;
    push (@flags, $CppFlags{$spath}) if(exists $CppFlags{$spath});
    my $dir_info = $obj->dir_info;
    push @flags, @{$dir_info->{sources_flags}};
    foreach my $hdr ($obj->used_headers)
    {
        push (@flags, $CppFlags{$hdr}) if(exists $CppFlags{$hdr});
    }
    my $rv;
    if(@flags)
    {
        $rv .= "$f: CPPFLAGS+=";
        $rv .= join(' ', @flags);
        $rv .= "\n";
    }
    return $rv;
}

sub add_linking_flags
{
    my $target = shift;
    my $obj_list = shift;
    # let's see if we need to add some ldflags or ldlibs to the target definition
    my @ldflags;
    my @ldlibs;
    foreach my $f (@$obj_list)
    {
        my $obj = $Objs{$f};
        my $spath = $obj->source_path;
        push (@ldlibs, $LdLibs{$spath}) if(exists $LdLibs{$spath});
        push (@ldflags, $LdFlags{$spath}) if(exists $LdFlags{$spath});
        foreach my $hdr ($obj->used_headers)
        {
            push (@ldlibs, $LdLibs{$hdr}) if(exists $LdLibs{$hdr});
            push (@ldflags, $LdFlags{$hdr}) if(exists $LdFlags{$hdr});
        }
    }
    my $rv;
    if(@ldflags)
    {
        $rv .= "$target: LDFLAGS+=";
        $rv .= join(" ", @ldflags);
        $rv .= "\n";
    }
    if(@ldlibs)
    {
        $rv .= "$target: LDLIBS+=";
        $rv .= join(" ", @ldlibs);
        $rv .= "\n";
    }
    return $rv;
}

sub load_directory
{
    my $dir = shift;
    return if exists $Dirs{$dir};
    warn("Processing directory '$dir'\n");
    $info = new_info();
    process_hints($dir, $info);
    $Dirs{$dir} = $info;
    # now I need to do some preprocessing of the information.
    # All of the paths are relative to that directory and not
    # to this one.
    if(@{$info->{sources}})
    {
        foreach my $f (@{$info->{sources}})
        {
            $f = compress_path("$dir/$f");
            die "$dir/$HintsFile: The entry '$f' from sources does not exist\n" unless -f $f || -d $f;
        }
    }
    else
    {
        @{$info->{sources}} = <$dir/*.c>;
    }
    if(@{$info->{headers}})
    {
        foreach my $f (@{$info->{headers}})
        {
            $f = compress_path("$dir/$f");
            die "$dir/$HintsFile: The entry '$f' from headers does not exist\n" unless -f $f || -d $f;
        }
    }
    else
    {
        @{$info->{headers}} = <$dir/*.h>;
    }
    foreach my $d (@{$info->{header_dirs}})
    {
        $d = compress_path("$dir/$d");
        die "$dir/$HintsFile: The headers directory $d does not exist\n" unless -d $d;
    }
    # I need to copy all of the flags into appropriate hash
    my $hash = delete $info->{ldflags};
    foreach my $key (keys %$hash)
    {
        my $f = compress_path("$dir/$key");
        die "$dir/$HintsFile: LDFLAGS are applied for non-existing file $f\n" unless -f $f;
        die "$dir/$HintsFile: duplicated LDFLAGS for file $f\n" if exists $LdFlags{$f};
        $LdFlags{$f} = $hash->{$key};
    }
    $hash = delete $info->{ldlibs};
    foreach my $key (keys %$hash)
    {
        my $f = compress_path("$dir/$key");
        die "$dir/$HintsFile: LDLIBS are applied for non-existing file $f\n" unless -f $f;
        die "$dir/$HintsFile: duplicated LDLIBS for file $f\n" if exists $LdLibs{$f};
        $LdLibs{$f} = $hash->{$key};
    }
    $hash = delete $info->{cppflags};
    foreach my $key (keys %$hash)
    {
        my $f = compress_path("$dir/$key");
        die "$dir/$HintsFile: CPPFLAGS are applied for non-existing file $f\n" unless -f $f;
        die "$dir/$HintsFile: duplicated CPPFLAGS for file $f\n" if exists $CppFlags{$f};
        $CppFlags{$f} = $hash->{$key};
    }
}


# this function attempts to create all directories in the path
# if the path does not contain any / then nothing is done
sub make_all_dirs
{
    my $path = shift;
    my $cwd = getcwd();
    my $need_jump_back = 0;

    while($path =~ s/^([^\/]+)\///)
    {
        my $d = $1;
        unless(-d $d)
        {
            mkdir($d) || die "Can't create directory $d in @{[getcwd]}: $!\n";
        }
        chdir($d);
        $need_jump_back = 1;
    }
    chdir($cwd) if $need_jump_back;
}

sub check_prog_or_lib
{
    my $args = shift;
    my $feat = shift;
    return "Missing argument 'name'" unless exists $args->{name};
    return "The 'name' can't start with / or be empty" if !$args->{name} || $args->{name} =~ /^\//;
    return "Missing argument '$feat'" unless exists $args->{$feat};
    # each element of the list has to point to directory or file.
    return "'objects' must be reference to array" unless ref $args->{$feat} eq 'ARRAY';
    foreach my $el (@{$args->{$feat}})
    {
        return "element in '$feat' can not have absolute path or be empty"
            if $el eq '' || substr($el, 0, 1) eq '/';
        next if -f $el;
        next if -d $el;
        return "'$feat' contains element '$el' which is not file or directory";
    }
    return;
}



sub new_info
{
    my %info;
    $info{sources} = [];
    $info{headers} = [];
    $info{header_dirs} = [];
    $info{ldflags} = {};
    $info{ldlibs} = {};
    $info{cppflags} = {};
    $info{sources_flags} = [];
    return \%info;
}

sub process_hints
{
    my $dir = shift;
    return unless(open(FILE, "$dir/$HintsFile"));
    my $info = shift;
    local($/) = undef;
    my $code = <FILE>;
    close(FILE);
    my $cwd = cwd();
    chdir($dir) || die "Can't enter $dir: $!\n";
    *wrapped::sources = $info->{sources};
    *wrapped::headers = $info->{headers};
    *wrapped::sources_flags = $info->{sources_flags};
    *wrapped::header_dirs = $info->{header_dirs};
    *wrapped::LDFLAGS = $info->{ldflags};
    *wrapped::LDLIBS = $info->{ldlibs};
    *wrapped::CPPFLAGS = $info->{cppflags};
    eval "package wrapped; $code";
    die "$dir/$HintsFile: $@\n" if $@;
    chdir($cwd) || die "Can't enter $cwd: $!\n";
}

sub optimize_path
{
    my $parts = shift;

    my $i = 0;
    while($i < @$parts)
    {
        # any '.' inside of the path is thrown away
        if($parts->[$i] eq '.')
        {
            splice(@$parts, $i, 1);
            next;
        }

        if($i != 0 && $parts->[$i] eq '..' && $parts->[$i-1] ne '..')
        {
            # the "/../" is equal to "/"
            if($i == 1 && $parts->[0] eq '')
            {
                splice(@$parts, 1, 1);
                next;
            }
            # the "dir/.." is equal to ""
            splice(@$parts, $i - 1, 2);
            $i--;
            next;
        }
        $i++;
    }
}

sub shift_path
{
    my $dst_path = shift;
    my $path = shift;
    return $path if($path =~ /^\//);

    my $cdir = shift;

    my @dp = split /\/+/, $dst_path;
    my @sp = split /\/+/, $path;
    my @cp = split /\/+/, $cdir;

    optimize_path(\@dp);
    optimize_path(\@sp);
    optimize_path(\@cp);

    # skip equal elements from the beginning. If the skipped element is '..' then
    # pop the element from current path. Otherwise push skipped element to current path
    while(@dp && @sp && $dp[0] eq $sp[0])
    {
        my $el = shift @dp;
        shift @sp;
        if($el eq '..') { pop(@cp); }
        else { push(@cp, $el); }
    }

    # if the first element of destination is '..' then I need to add directories from
    # current path
    while(@dp && $dp[0] eq '..')
    {
        shift @dp;
        unshift @sp, pop @cp if @cp;
    }

    # now for each element in @dp I need to add '..' to @sp
    while(@dp)
    {
        shift @dp;
        unshift @sp, '..';
    }

    unshift @sp, '.' if(@sp == 0 || ($sp[0] ne '' && $sp[0] ne '..'));
    return join('/', @sp);
}

sub compress_path
{
    my $path = shift;
    return '' if $path eq '';
    my @parts = split /\/+/, $path;
    return '/' unless @parts; # the split produces empty list for "/"!!

    optimize_path(\@parts);
    # I want all relative paths to start with '.'
    unshift @parts, '.' if(@parts == 0 || ($parts[0] ne '' && $parts[0] ne '..'));
    return join('/', @parts);
}

sub wrapped::runs_fine
{
    my $info = pop;

    my $cwd = getcwd();
    my $chk;
    my $dir = File::Temp->newdir();
    chdir($dir->dirname) || die "Can't go into tmp dir: $!\n";

    my $base = 'checker';
    open(FILE, ">$base.c") || die "Can't create checker.c: $!\n";
    print FILE $info->{code};
    close(FILE);

    open(OLDOUT, ">&STDOUT") || die "Can't dup STDOUT: $!\n";
    open(OLDERR, ">&STDERR") || die "Can't dup STDERR: $!\n";
    open(STDOUT, ">/dev/null") || die "Can't redirect STDOUT: $!\n";
    open(STDERR, ">/dev/null") || die "Can't redirect STDERR: $!\n";

    my @cmd = qw(gcc);
    push(@cmd, $info->{cppflags}) if exists $info->{cppflags};
    push(@cmd, qw(-c), "$base.c");
    $chk = system @cmd;
    goto cleanup if($chk != 0);

    @cmd = qw(gcc);
    push(@cmd, $info->{ldflags}) if exists $info->{ldflags};
    push(@cmd, qw(-o), $base, "$base.o");
    push(@cmd, $info->{ldlibs}) if exists $info->{ldlibs};
    $chk = system @cmd;
    goto cleanup if($chk != 0);

    $chk = system "./$base";

cleanup:
    open(STDOUT, ">&OLDOUT") || die "Can't dup STDOUT: $!\n";
    open(STDERR, ">&OLDERR") || die "Can't dup STDERR: $!\n";
    close(OLDERR);
    close(OLDOUT);
    chdir($cwd);
    return $chk == 0;
}

package ObjHolder;

sub new
{
    my $pkg = shift;
    my $spath = shift;
    my $dir_info = shift;
    return bless [$spath, $dir_info, []], $pkg;
}

sub source_path
{
    my $this = shift;
    return $this->[0];
}

sub set_headers
{
    my $this = shift;
    $this->[3] = shift;
}

sub used_headers
{
    my $this = shift;
    return @{$this->[3]};
}

sub add_header_dir
{
    my $this = shift;
    push(@{$this->[2]}, @_);
}

sub compile_flags
{
    my $this = shift;
    return map {"-I$_"} @{$this->[2]};
}

sub dir_info
{
    my $this = shift;
    return $this->[1];
}
