/* this is an implementation of simple echo server. This server accepts
 * connections on localhost:6080 and simply prints on stdout whatever it reads.
 * If there was nothing written within 1 minute, then connection is closed.
 * After accepting 10 connections the server finishes its work.
 * Start reading from the bottom function.
 */

#include "evtid.h"
#include "predok.h"
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdarg.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <stdio.h>

typedef struct peer_ctrl_t peer_ctrl_t;
typedef struct lstn_ctrl_t lstn_ctrl_t;

struct peer_ctrl_t{
    evd_t evd;
    evti_t tmr;
};

struct lstn_ctrl_t{
    evd_t evd;
    int count;
};

static void
error(int e_code, int err, const char * fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    vfprintf(stderr, fmt, ap);
    if(err)
	fprintf(stderr, ": %s (%d)", strerror(err), err);
    fputs("\n", stderr);
    fflush(stderr);
    if(e_code) exit(e_code);
}

static void
set_non_block(int fd)
{
    if(fcntl(fd, F_SETFL, O_NONBLOCK) != 0) 
	error(EXIT_FAILURE, errno, "Can't set non-blocking");
}

static void
unbelivable(evd_t * evd)
{
    error(EXIT_FAILURE, errno, "We shouldn't have requested readinnes for writing?");
}

static void
release_peer(evd_t * evd)
{
    peer_ctrl_t * ctrl = predok(evd, peer_ctrl_t, evd);
    evd_nowatch(evd);
    evti_stop(&ctrl->tmr);
    close(evd->d);
    free(ctrl);
}

static void
reset_tmr(evd_t * evd)
{
    peer_ctrl_t * ctrl = predok(evd, peer_ctrl_t, evd);
    evti_stop(&ctrl->tmr);
    evti_start_s(&ctrl->tmr, 60);
}

static void
cnx_error(evd_t * evd)
{
    error(0, errno, "Closing connection on %d after an error", evd->d);
    release_peer(evd);
}

static void
read_input(evd_t * evd)
{
    char buf[64];
    ssize_t chk;

    chk = recv(evd->d, buf, sizeof(buf), 0);
    if(chk < 0)
    {
        if(errno == EAGAIN || errno == EWOULDBLOCK)
        {
            evd_input_exhausted(evd);
            return;
        }
        cnx_error(evd);
    }
    if(chk == 0) 
    {
	printf("Sock %d EOF\n", evd->d);
	release_peer(evd);
    }
    else
    {
	printf("Sock %d:--", evd->d);
	fwrite(buf, chk, 1, stdout);
	fputs("--\n", stdout);
	// restart the timer so that it waits one minute after this moment of
	// time
	reset_tmr(evd);
	// since the evd_read was requested, this function will be called
	// again if there's some data in the buffer.
    }
}


// this function is called if there was no input on some connection during 1
// minute. Here we simply close the connection and release all resources
static void
close_cnx(evti_t * tmr)
{
    peer_ctrl_t * ctrl = predok(tmr, peer_ctrl_t, tmr);
    printf("Connection on socket %d no input within 1 minute\n", ctrl->evd.d);
    evd_nowatch(&ctrl->evd);
    close(ctrl->evd.d);
    free(ctrl);
}

// this is read handler for listener. We attempt to accept new connection
// here.
static void
accept_cnx(evd_t * lstn)
{
    int peer;
    lstn_ctrl_t * lc;
    peer_ctrl_t * ctrl;

    peer = accept(lstn->d, NULL, NULL);
    if(peer < 0)
    {
        if(errno != EAGAIN && errno != EWOULDBLOCK) error(EXIT_FAILURE, errno, "error in accept");
        evd_input_exhausted(lstn);
        return;
    }

    set_non_block(peer);

    ctrl = calloc(1, sizeof(*ctrl));
    if(ctrl == NULL) error(EXIT_FAILURE, errno, "no memory");

    evd_init(&ctrl->evd);
    ctrl->evd.d = peer;
    ctrl->evd.raction = read_input;
    ctrl->evd.waction = unbelivable;
    evd_watch(&ctrl->evd);
    evd_raction_active(&ctrl->evd);

    evti_init(&ctrl->tmr);
    ctrl->tmr.action = close_cnx;
    // expire connection after 60 seconds
    evti_start_s(&ctrl->tmr, 60);

    // check how many connections we have received
    // if already 10, then close listener
    lc = (lstn_ctrl_t *)lstn;
    lc->count++;
    if(lc->count == 10)
    {
	evd_nowatch(lstn);
	close(lstn->d);
    }
}

int
main(void)
{
    struct sockaddr_in myadr;
    lstn_ctrl_t lstn = {{0}};

    evtid_init();

    myadr.sin_family = AF_INET;
    myadr.sin_port = htons(6080);
    inet_aton("127.0.0.1", &myadr.sin_addr);

    evd_init(&lstn.evd);
    lstn.evd.d = socket(PF_INET, SOCK_STREAM, 0);
    if(lstn.evd.d < 0) error(EXIT_FAILURE, errno, "No socket");
    if(bind(lstn.evd.d, (struct sockaddr*)&myadr, sizeof(myadr)) != 0)
	error(EXIT_FAILURE, errno, "Can't bind");
    listen(lstn.evd.d, 10);

    set_non_block(lstn.evd.d);

    lstn.evd.raction = accept_cnx;
    lstn.evd.waction = unbelivable;
    evd_watch(&lstn.evd);
    evd_raction_active(&lstn.evd);

    evtid_dispatcher();
    return 0;
}
