/* this is an implementation of "persistant" tail -f command. This command
 * shall read the lines from the end of the named file. It shall notice if the
 * file was moved or removed and switch back to reading of the named file.
 * Beware. It'll create the file if it does not exist. This program is
 * convient for watching the log file that might get rotated.
 *
 * This program is known to work on Linux and on NetBSD. 
 */

#include "evtid.h"
#include "predok.h"
#include "tail_f.h"
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <fcntl.h>

#ifndef SLEEP_TIME
#define SLEEP_TIME 500
#endif

typedef struct file_info_t file_info_t;

struct file_info_t{
    evti_t tmr;
    FILE * fs;
    char * name;
    dev_t st_dev;
    ino_t st_ino;
};

static void
error(int e_code, int err, const char * fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    vfprintf(stderr, fmt, ap);
    if(err)
	fprintf(stderr, ": %s (%d)", strerror(err), err);
    fputs("\n", stderr);
    fflush(stderr);
    if(e_code) exit(e_code);
}

static int
do_getline(FILE * fs, char * line)
{
    int chr;
    int len;
    chr = fgetc(fs);
    if(chr == EOF) return EOF;
    if(chr == '\n')
    {
        line[0] = '\n';
        line[1] = 0;
        return 1;
    }
    ungetc(chr, fs);
    fscanf(fs, "%4000[^\n]%n", line, &len);
    // now make sure that next character is new line
    chr = fgetc(fs);
    if(chr == '\n' || chr == EOF)
    {
        line[len++] = '\n';
        line[len] = 0;
    }
    else ungetc(chr, fs);
    return len;
}

static FILE *
my_fopen(const char * fname)
{
    int fd = open(fname, O_RDONLY|O_CREAT, 0666);
    if(fd < 0) return NULL;
    return fdopen(fd, "r");
}

static void
read_file(evti_t * tmr)
{
    char line[4096];
    int len;
    file_info_t * finfo = predok(tmr, file_info_t, tmr);
    len = do_getline(finfo->fs, line);
    // if we have reached the end of file, we check if the file was removed,
    // or moved or anything else happened to it. Then we read new file from
    // the beginning.
    if(len == EOF)
    {
	struct stat ss;
	clearerr(finfo->fs);
	len = stat(finfo->name, &ss);
	if(len != 0 
		|| ss.st_dev != finfo->st_dev 
		|| ss.st_ino != finfo->st_ino
		|| ss.st_size < ftell(finfo->fs))
	{
	    fclose(finfo->fs);
	    finfo->fs = my_fopen(finfo->name);
	    if(finfo->fs == NULL) error(EXIT_FAILURE, errno, "Can't access %s", finfo->name);
	    if(fstat(fileno(finfo->fs), &ss) != 0) 
		error(EXIT_FAILURE, errno, "Can't reaccess %s", finfo->name);
	    finfo->st_dev = ss.st_dev;
	    finfo->st_ino = ss.st_ino;
	    evti_start(tmr, 0);
	}
	else
	    evti_start(tmr, SLEEP_TIME);
    }
    else
    {
	// if we got the line, then we simply print it and then sleep for 0
	// seconds and do it again
	fputs(line, stdout);
	fflush(stdout);
	evti_start(tmr, 0);
    }
}

int
main(int argc, char * argv[])
{
    file_info_t finfo;

    if(argc == 1)
	error(EXIT_FAILURE, 0, "Usage: %s <file_name>", argv[0]);
    evtid_init();
    finfo.name = argv[1];
    finfo.fs = my_fopen(finfo.name);
    if(finfo.fs == NULL) error(EXIT_FAILURE, errno, "Can't read '%s'", finfo.name);
    fseek(finfo.fs, 0, SEEK_END);
    {
	struct stat ss;
	if(fstat(fileno(finfo.fs), &ss) != 0) 
	    error(EXIT_FAILURE, errno, "Can't access %s", finfo.name);
	finfo.st_dev = ss.st_dev;
	finfo.st_ino = ss.st_ino;
    }
    evti_init(&finfo.tmr);
    finfo.tmr.action = read_file;
    evti_start(&finfo.tmr, SLEEP_TIME);
    evtid_dispatcher();
    evtid_fini();
    return 0;
}
