/* this program sets up random number of timers and upon expiration of those
 * makes sure how much time has actually expired. It uses ACCURACY parameter
 * defined below. Currently this is set to 50 ms. But this may depend on the
 * hardware and system load, so possibly higher values may be required. It
 * still shouldn't be more than 100. At the end the program prints out average
 * delay with execution of timer
 */

#include "evtid.h"
#include <stdarg.h>
#include <errno.h>
#include <stdlib.h>
#include <sys/time.h>
#include <stdio.h>
#include <string.h>
#include <time.h>

static void
error(int e_code, int err, const char * fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    vfprintf(stderr, fmt, ap);
    if(err)
	fprintf(stderr, ": %s (%d)", strerror(err), err);
    fputs("\n", stderr);
    fflush(stderr);
    if(e_code) exit(e_code);
}


typedef struct my_tmr{
    evti_t tmr;
    uint64_t limit;
#ifdef USE_TIME
    struct timeval time;
#else
    evti_mark_t time;
#endif
} my_tmr;

static size_t idx;
#define ACCURACY 80

static int total_errors;
static int total_tmrs;
static uint64_t off;


#ifdef USE_TIME
static uint64_t
time_diff(struct timeval *l, struct timeval *r)
{
    register uint64_t sec = l->tv_sec - r->tv_sec;
    register long usec;
    if(l->tv_usec < r->tv_usec)
    {
	usec = 1000000 + l->tv_usec - r->tv_usec;
	sec--;
    }
    else
	usec = l->tv_usec - r->tv_usec;
    return sec * 1000 + usec / 1000;
}
#endif

static void
add_timer(int delay, void (*h)(evti_t*))
{
    my_tmr * ctrl;
    ctrl = malloc(sizeof(*ctrl));
    if(ctrl == NULL) error(EXIT_FAILURE, errno, "Can't start timer for %d ms", delay);
    evti_init(&ctrl->tmr);
    ctrl->tmr.action = h;
    ctrl->limit = delay;
#ifdef USE_TIME
    gettimeofday(&ctrl->time, NULL);
#else
    evti_delay_mark(&ctrl->time);
#endif
    evti_start(&ctrl->tmr, ctrl->limit);
#ifdef DEBUG_TIMER
    fprintf(stderr, "Added timer %p for %d ms\n", ctrl, delay);
#endif
}

static void
check_acuracy(evti_t * tmr)
{
    my_tmr * ctrl = (my_tmr*)tmr;
    uint64_t diff;
    uint64_t delta;
#ifdef USE_TIME
    {
	struct timeval now;
	gettimeofday(&now, NULL);
	diff = time_diff(&now, &ctrl->time);
    }
#else
    diff = evti_delay(&ctrl->time);
#endif

    delta = (diff > ctrl->limit) ? diff - ctrl->limit : ctrl->limit - diff;

    if(delta > ACCURACY) 
    {
	error(0, 0, "WARNING: %p Expected %lu got %lu", tmr,
		(unsigned long) ctrl->limit, (unsigned long)diff);
	total_errors++;
    }


    total_tmrs++;
    off += delta;

#ifdef DEBUG_TIMER
    fprintf(stderr, "After delay %lu, delta %lu\n", 
	    (unsigned long) ctrl->limit, (unsigned long) delta);
#endif

}

static void
handler(evti_t * tmr)
{
    check_acuracy(tmr);
    if(idx != 0)
    {
	int count = rand();
	count &= 0xf;
	count++;
	do
	{
	    int delay = rand();
	    delay &= 0xffff;
	    add_timer(delay, handler);
	    count--;
	}while(count != 0);
	idx--;
    }
    free(tmr);
}
/*
static void
simple_handler(evti_t * tmr)
{
    check_acuracy(tmr);
    free(tmr);
}
*/

static void
tricky_handler(evti_t * tmr)
{
    check_acuracy(tmr);
    free(tmr);
    add_timer(4, handler);
}

int
main(int argc, char * argv[])
{
    long seed;
    if(argc > 1)
	seed = strtol(argv[1], NULL, 10);
    else
	seed = time(0);
    fprintf(stderr, "Using seed %ld\n", seed);
    srand(seed);
    evtid_init();

    add_timer(1024, handler);
    add_timer(60, tricky_handler);
    add_timer(65534, handler);

    // generate random number of timers with random delay
    /*
    idx = rand();
    idx &= 0xf;
    */
    idx = 100000;
    evtid_dispatcher();
    evtid_fini();
    printf("Average offset is %d\n", (int)(off / total_tmrs));
    printf("Total number of WARNINGs is %d\n", total_errors);
    return 0;
}
